#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Entry point into the IDMS Linux installer from the commandline."""

import argparse
import logging
import logging.handlers
import os
import sys
from typing import Any, MutableMapping, Optional

import jinja2
import requests
import toml

from ilitools import IliBase, util

from .target.container_dir import IliTargetContainerDir
from .target.container_tar import IliTargetContainerTar
from .target.container_tar_docker import IliTargetContainerTarDocker
from .target.disk_image import IliTargetDiskImage
from .target.disk_install import IliTargetDiskInstall
from .target.live_image import IliTargetLiveImage
from .version import __version__

__all__ = ["run", "run_ili"]


def add_create_arguments(arg_group: argparse.ArgumentParser) -> None:
    """Add global arguments."""

    arg_group.add_argument("--profile", nargs=1, help="Target profile config file")
    arg_group.add_argument(
        "--data-search-paths",
        nargs=1,
        help="Search paths for ilitools data (comma separated)",
    )
    arg_group.add_argument("--staging", action="store_true", help="Use staging repositories")
    arg_group.add_argument("--pacman-conf", nargs=1, help="Use a specific pacman configuration file")
    arg_group.add_argument("--packages", nargs=1, help="Additional packages to install (comma separated)")
    arg_group.add_argument("--services", nargs=1, help="Additional services to enable (comma separated)")
    arg_group.add_argument(
        "--hostname",
        nargs=1,
        help="Set system hostname (all targets default to localhost.localdomain)",
    )
    arg_group.add_argument(
        "--overlay-directory",
        nargs=1,
        help="Add an overlay directory in the format of SRC:DST",
    )
    arg_group.add_argument(
        "--enable-root-autologin",
        dest="root_autologin",
        action="store_true",
        default=None,
        help="Setup root auto-login (enabled by default for live-image)",
    )
    arg_group.add_argument(
        "--disable-root-autologin",
        dest="root_autologin",
        action="store_false",
        default=None,
        help="Disable root auto-login (enabled by default for live-image)",
    )
    arg_group.add_argument(
        "--enable-install-microcode",
        dest="install_microcode",
        action="store_true",
        default=None,
        help="Enable installation of CPU microcode updates (enabled by default for disk-install and live-image)",
    )
    arg_group.add_argument(
        "--disable-install-microcode",
        dest="install_microcode",
        action="store_false",
        default=None,
        help="Disable installation of CPU microcode updates (enabled by default for disk-install and live-image)",
    )
    arg_group.add_argument(
        "--disable-networking",
        dest="networking",
        action="store_false",
        default=None,
        help="Disable networking",
    )
    arg_group.add_argument(
        "--network-devices",
        nargs=1,
        help="Network devices to configure (comma separated)",
    )


def add_create_container_dir_arguments(arg_group: argparse.ArgumentParser) -> None:
    """Add container-dir target arguments."""
    arg_group.add_argument("--target-dir", nargs=1, required=True, help="Target directory")


def add_create_container_tar_arguments(arg_group: argparse.ArgumentParser) -> None:
    """Add create container tar arguments."""
    arg_group.add_argument(
        "--target-file",
        nargs=1,
        metavar="TARGET_TAR_FILE",
        required=True,
        help="Target file (.tar.{gz,bz2,xz,zst)",
    )


def add_create_disk_image_arguments(arg_group: argparse.ArgumentParser) -> None:
    """Add create disk image arguments."""
    arg_group.add_argument("--target-image", nargs=1, required=True, help="Target image file")
    arg_group.add_argument("--image-size", nargs=1, help="Target image size (>= 2G, defaults to 4G)")
    arg_group.add_argument("--disk-label", nargs=1, help="Disk label (eg. [gpt] or dos)")
    # Options from disk_generic
    arg_group.add_argument(
        "--enable-efi",
        dest="create_efi",
        action="store_true",
        default=None,
        help="Enable EFI booting",
    )
    arg_group.add_argument(
        "--enable-bootfs",
        dest="create_bootfs",
        action="store_true",
        default=None,
        help="Create separate /boot filesystem",
    )
    arg_group.add_argument(
        "--rootfs-fstype",
        nargs=1,
        help="Filesystem type to use for the rootfs (eg. [ext4], btrfs)",
    )
    arg_group.add_argument(
        "--disable-autoresize-root",
        dest="autoresize_root",
        action="store_false",
        default=None,
        help="Disable autoresizing of root volume",
    )
    arg_group.add_argument(
        "--enable-microcode-update",
        dest="install_microcode",
        action="store_true",
        default=None,
        help="Enable CPU microcode updates",
    )
    arg_group.add_argument(
        "--kernel",
        nargs=1,
        help="Kernel to use (eg. linux, [linux-lts], linux-hardened)",
    )
    arg_group.add_argument("--serial-terminal", nargs=1, help="Enable serial terminal (eg. ttyS0,115200)")
    arg_group.add_argument(
        "--serial-terminal-type",
        nargs=1,
        help="Set specific serial terminal type. (eg. vt100, linux)",
    )
    arg_group.add_argument(
        "--mkinitcpio-modules",
        nargs=1,
        help="List of modules to add to mkinitcpio (eg. ata,piix)",
    )
    arg_group.add_argument(
        "--enable-mkinitcpio-autodetect",
        dest="mkinitcpio_autodetect",
        action="store_true",
        default=None,
        help="Try determine what kernel modules we need, instead of adding all of them",
    )
    arg_group.add_argument(
        "--disable-cloudinit",
        dest="use_cloudinit",
        action="store_false",
        default=None,
        help="Disable cloud-init",
    )
    arg_group.add_argument(
        "--cloudinit-enable-root",
        action="store_true",
        default=None,
        help="Enable root login when using cloud-init (disabled by default)",
    )
    arg_group.add_argument(
        "--cloudinit-admin-username",
        nargs=1,
        help="Set the admin username created by cloud-init default configuration",
    )

    arg_group.add_argument(
        "--sshd-client-alive-interval",
        nargs=1,
        help="Set the sshd ClientAliveInterval option",
    )


def add_create_disk_install_arguments(arg_group: argparse.ArgumentParser) -> None:
    """Add create disk install arguments."""

    arg_group.add_argument("--target-disks", nargs=1, help="Target disks (comma separated)")
    arg_group.add_argument(
        "--disk-usage-strategy",
        nargs=1,
        help="Disk usage strategy (eg. ASIS, MDRAID1, MDRAID5, MDRAID6, MDRAID10)",
    )
    arg_group.add_argument("--disk-layout-strategy", nargs=1, help="Disk usage strategy (eg. FLAT, LVM)")
    arg_group.add_argument(
        "--disk-encryption-strategy", nargs=1, help="Disk encryption strategy (eg. NONE, LUKS-ON-PART, LVM-ON-LUKS)"
    )
    arg_group.add_argument(
        "--network-device-naming-strategy",
        nargs=1,
        help="Network device naming strategy (eg. ASIS, AUTO-SEQ)",
    )

    arg_group.add_argument(
        "--disable-efi",
        dest="create_efi",
        action="store_false",
        default=None,
        help="Disable EFI booting",
    )
    arg_group.add_argument(
        "--disable-bootfs",
        dest="create_bootfs",
        action="store_false",
        default=None,
        help="Do not create /boot filesystem",
    )
    arg_group.add_argument(
        "--rootfs-fstype",
        nargs=1,
        help="Filesystem type to use for the rootfs (eg. [ext4], btrfs)",
    )
    arg_group.add_argument(
        "--enable-autoresize-root",
        dest="autoresize_root",
        action="store_true",
        default=None,
        help="Enable autoresizing of rootfs",
    )
    arg_group.add_argument("--kernel", nargs=1, help="Kernel to use (eg. linux, linux-lts, linux-hardened)")
    arg_group.add_argument("--serial-terminal", nargs=1, help="Enable serial terminal (eg. ttyS0,115200)")
    arg_group.add_argument(
        "--serial-terminal-type",
        nargs=1,
        help="Set specific serial terminal type. (eg. vt100, linux)",
    )
    arg_group.add_argument(
        "--mkinitcpio-modules",
        nargs=1,
        help="List of modules to add to mkinitcpio (eg. ata,piix)",
    )
    arg_group.add_argument(
        "--disable-mkinitcpio-autodetect",
        dest="mkinitcpio_autodetect",
        action="store_false",
        default=None,
        help="Do not try determine what kernel modules we need, add all of them",
    )
    arg_group.add_argument(
        "--enable-cloudinit",
        dest="use_cloudinit",
        action="store_true",
        default=None,
        help="Enable cloud-init",
    )
    arg_group.add_argument(
        "--cloudinit-enable-root",
        action="store_true",
        default=None,
        help="Enable root login when using cloud-init (disabled by default)",
    )
    arg_group.add_argument(
        "--cloudinit-admin-username",
        nargs=1,
        help="Set the admin username created by cloud-init default configuration",
    )
    arg_group.add_argument(
        "--sshd-client-alive-interval",
        nargs=1,
        help="Set the sshd ClientAliveInterval option",
    )


def add_create_live_image_arguments(arg_group: argparse.ArgumentParser) -> None:
    """Add create live image arguments."""
    arg_group.add_argument("--target-file", nargs=1, required=True, help="Target file (.img)")
    arg_group.add_argument("--autorun", nargs=1, help="Automatically run this script on bootup")
    arg_group.add_argument(
        "--ilitools-profile",
        nargs=1,
        help="Pre-seed the ilitools profile to use on the kernel command line",
    )


def get_argparser() -> argparse.ArgumentParser:
    """Return an instance of argparse.ArgumentParser."""

    print(f"IDMS Linux Installer v{__version__} - Copyright © 2019-2024, AllWorldIT.\n", file=sys.stderr)

    argparser = argparse.ArgumentParser(add_help=False)

    optional_group = argparser.add_argument_group("Optional arguments")
    optional_group.add_argument("-h", "--help", action="help", help="Show this help message and exit")
    optional_group.add_argument("-v", "--verbose", action="store_true", help="Display verbose logging")

    return argparser


def setup_logging(args: argparse.Namespace) -> None:
    """Set up logging."""

    # Setup logger and level
    logger = logging.getLogger()
    if args.verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    # Setup console handler
    console_handler = logging.StreamHandler()
    # Use a better format for messages
    console_handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)-8s %(message)s"))
    logger.addHandler(console_handler)


def create_get_profile(
    parsed_args: argparse.Namespace,
) -> Optional[MutableMapping[str, Any]]:
    """Return a parsed profile."""

    # Check if we have a profile specified
    target_profile_path: Optional[str] = None
    if parsed_args.profile:
        target_profile_path = parsed_args.profile[0]
    # If we don't have a profile set, check the kernel commandline
    else:
        # Grab kernel command line
        kernel_cmdline = util.kernel_cmdline()
        # Check if we have a profile specified
        if "ilitools.profile" in kernel_cmdline:
            # Check it has a value...
            if kernel_cmdline["ilitools.profile"]:
                # If it does we can probably use it
                target_profile_path = kernel_cmdline["ilitools.profile"]
                logging.info("Using profile from kernel commandline: %s", target_profile_path)
            # If there is no value we need to throw an error
            else:
                logging.error("Kernel commandline option 'ilitools.profile' specified without a value")
                sys.exit(1)

    # If we don't have a profile here, we can just return
    if not target_profile_path:
        return None

    # Check if we have a profile configuration
    target_profile: Optional[MutableMapping[str, Any]] = None

    # Check if its something we need to download
    if target_profile_path.startswith("http"):
        logging.info("Using profile: %s", target_profile_path)
        # Grab HTTP resource
        try:
            resource = requests.get(target_profile_path, headers={"User-Agent": f"IliTools/{__version__}"}, timeout=30)
            resource.raise_for_status()
        except requests.exceptions.HTTPError as err:
            logging.error("Error loading profile from '%s': %s", target_profile_path, err)
            sys.exit(1)
        # Make sure we got something back
        if resource.text:
            # Parse the toml we got
            target_profile = toml.loads(resource.text)
        else:
            logging.error("Error loading profile from '%s': No data returned", target_profile_path)
            sys.exit(1)

    # In this case its on the local filesystem so we can use templating...
    else:
        # Open the profile
        target_profile_fname = os.path.basename(target_profile_path)
        target_profile_dir = os.path.dirname(target_profile_path)

        search_paths = []
        # Add target profile base directory
        search_paths.append(target_profile_dir)
        # Add system default location
        search_paths.extend(util.search_paths(parsed_args.data_search_paths, "profiles"))
        logging.debug("Using profile search paths: %s", search_paths)

        # Render first with jinja
        template_env = jinja2.Environment(  # nosec
            loader=jinja2.FileSystemLoader(searchpath=search_paths),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = template_env.get_template(target_profile_fname)
        # Then parse the toml
        target_profile = toml.loads(template.render())

    logging.info("Using profile: %s", target_profile_path)

    return target_profile


def run() -> None:  # noqa: CFQ001 # pylint: disable=too-many-statements
    """Entry point for execution from the commandline."""

    # Grab our arg parser
    argparser = get_argparser()

    subparsers = argparser.add_subparsers(help="Action to perform")

    # CMD: create
    parser_create = subparsers.add_parser("create", help="Create an installation")
    # Set action
    parser_create.add_argument(
        "--action",
        action="store_const",
        const="create",
        default="create",
        help=argparse.SUPPRESS,
    )
    # Add arguments
    add_create_arguments(parser_create)

    # CMD: create => *
    create_subparsers = parser_create.add_subparsers(help="Create target")

    # CMD: create container-dir
    parser_create_container_dir = create_subparsers.add_parser("container-dir", help="Create container in directory")
    # Set create-target
    parser_create_container_dir.add_argument(
        "--create-target",
        action="store_const",
        const="container-dir",
        default="container-dir",
        help=argparse.SUPPRESS,
    )
    # Add arguments
    add_create_container_dir_arguments(parser_create_container_dir)

    # CMD: create container-tar
    parser_create_container_tar = create_subparsers.add_parser("container-tar", help="Create container tar")
    parser_create_container_tar.add_argument(
        "--create-target",
        action="store_const",
        const="container-tar",
        default="container-tar",
        help=argparse.SUPPRESS,
    )
    # Add arguments
    add_create_container_tar_arguments(parser_create_container_tar)

    # CMD: create container-tar-docker
    parser_create_container_tar_docker = create_subparsers.add_parser("container-tar-docker", help="Create container tar docker")
    parser_create_container_tar_docker.add_argument(
        "--create-target",
        action="store_const",
        const="container-tar-docker",
        default="container-tar-docker",
        help=argparse.SUPPRESS,
    )
    # Add arguments
    add_create_container_tar_arguments(parser_create_container_tar_docker)

    # CMD: create disk-image
    parser_create_disk_image = create_subparsers.add_parser("disk-image", help="Create disk image")
    parser_create_disk_image.add_argument(
        "--create-target",
        action="store_const",
        const="disk-image",
        default="disk-image",
        help=argparse.SUPPRESS,
    )
    # Add arguments
    add_create_disk_image_arguments(parser_create_disk_image)

    # CMD: create disk-install
    parser_create_disk_install = create_subparsers.add_parser("disk-install", help="Create disk installation")
    parser_create_disk_install.add_argument(
        "--create-target",
        action="store_const",
        const="disk-install",
        default="disk-install",
        help=argparse.SUPPRESS,
    )
    # Add arguments
    add_create_disk_install_arguments(parser_create_disk_install)

    # CMD: create live-image
    parser_create_live_image = create_subparsers.add_parser("live-image", help="Create a live image")
    parser_create_live_image.add_argument(
        "--create-target",
        action="store_const",
        const="live-image",
        default="live-image",
        help=argparse.SUPPRESS,
    )
    # Add arguments
    add_create_live_image_arguments(parser_create_live_image)

    # Parse args
    args = argparser.parse_args()

    # If there is no action specified, throw the help at the user
    if "action" not in args:
        print("ERROR: No action specified!")
        argparser.print_help()
        sys.exit(1)

    setup_logging(args)

    # If its a create action, check what we're doing
    if args.action == "create":
        ilitools_create_target(parser_create, args)


def ilitools_create_target(subparser: argparse.ArgumentParser, args: argparse.Namespace) -> None:
    """Create a target."""

    # Check if we have a create_target attribute or not
    if hasattr(args, "create_target"):
        target = args.create_target
    # If the target is not specified in either, display an error
    else:
        logging.error("Target not specified on command line, please see help")
        subparser.print_help()
        sys.exit(1)

    logging.info("Creating target: %s", target)

    # Grab target profile
    target_profile = create_get_profile(args)

    # Go through each type of target we support
    ili: Optional[IliBase] = None
    if target == "container-dir":
        ili = IliTargetContainerDir(args, target_profile)
    elif target == "container-tar":
        ili = IliTargetContainerTar(args, target_profile)
    elif target == "container-tar-docker":
        ili = IliTargetContainerTarDocker(args, target_profile)
    elif target == "disk-image":
        ili = IliTargetDiskImage(args, target_profile)
    elif target == "disk-install":
        ili = IliTargetDiskInstall(args, target_profile)
    elif target == "live-image":
        ili = IliTargetLiveImage(args, target_profile)

    if not ili:
        logging.error("Unknown target '%s'", target)
        subparser.print_help()
        sys.exit(1)

    # Run the installer
    ili.run()


def run_ili() -> None:
    """Run our installer (disk-install)."""
    # Grab our arg parser
    argparser = get_argparser()
    # Add arguments
    add_create_disk_install_arguments(argparser)
    add_create_arguments(argparser)
    # Parse args
    args = argparser.parse_args()
    # Setup logging
    setup_logging(args)

    # Grab target profile
    target_profile = create_get_profile(args)

    # Start installer
    ili = IliTargetDiskInstall(args, target_profile)
    ili.run()


if __name__ == "__main__":
    run()
