#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Copyright (c) 2019-2024, AllWorldIT
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=too-many-lines

"""Disk installation target."""

import json
import logging
import os
import pathlib
import re
import tempfile
import uuid
from typing import Any, Dict, List, Optional, Tuple

from .. import util
from ..exceptions import IliException
from ..filesystem import IliFilesystem
from .disk_generic import IliTargetDiskGeneric

__all__ = ["IliTargetDiskInstall"]


class IliTargetDiskInstall(IliTargetDiskGeneric):  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Ili disk installation target."""

    # Available block devices
    _available_blockdevices_info: List[Tuple[str, str]]
    # Available network devices
    _available_network_devices: Dict[str, Dict[str, str]]

    # Target disks we're going to be using
    _target_disks: Optional[List[str]]
    # Disk strategies
    _target_disk_usage_strategy: Optional[str]
    _target_disk_layout_strategy: Optional[str]
    _target_disk_encryption_strategy: Optional[str]
    # Networking properties
    _target_network_device_naming_strategy: Optional[str]

    # LVM-specific, root filesystem size
    _target_lvm_rootfs_size: Optional[str]

    # Disk encryption password
    _target_disk_encryption_password: Optional[str]

    # Mount path of root filesystem
    _rootfs_path: Optional[str]

    # EFI volumes
    _efi_volumes: List[str]
    # Boot volume
    _boot_volume: Optional[str]
    # Main volume
    _main_volume: Optional[str]

    def target_init(self) -> None:
        """
        Initialize the target.

        This is part of the __init__ process.
        """
        # Hand over to parent
        super().target_init()

        # Setup available block devices
        self._available_blockdevices_info = self._get_available_blockdevices_info()

        # Setup available network devices
        self._available_network_devices = self._get_available_network_devices()

        # Target disks to install on
        self._target_disks = None
        # Disk strategies
        self._target_disk_usage_strategy = None
        self._target_disk_layout_strategy = None
        self._target_disk_encryption_strategy = None
        # Network device properties
        self._target_network_device_naming_strategy = None

        # LVM-specific, root filesystem size
        self._target_lvm_rootfs_size = "20G"

        # Target disk encryption password
        self._target_disk_encryption_password = None

        # Reset the rootfs fstype so we get a prompt
        self._target_rootfs_fstype = None

        # Default EFI size to 512M for disk installations
        self._target_efifs_size = "512M"

        # Mount path of root filesystem
        self._rootfs_path = None

        # EFI volumes
        self._efi_volumes = []
        # Boot volume
        self._boot_volume = None
        # Main volume
        self._main_volume = None

        # Set our defaults
        self.target_create_efi = True
        self.target_create_bootfs = True
        self.target_install_microcode = True

        # Add in stage to check and write out /etc/mdadm.conf if we're using software RAID
        self.add_stage(7900, "Setup /etc/mdadm.conf", self.setup_target_mdadm_conf)

    def prepare_configure(self) -> None:
        """Set up items we need for configuration."""

        self.add_required_config(
            2100,
            {
                "name": "target_disks",
                "type": "checkboxlist",
                "property": "target_disks",
                "required": True,
                "title": "Disk Selection",
                "prompt": "Disks to install on",
                "values": self.available_blockdevices_info,
                "multivalue": True,
            },
        )

        self.add_required_config(
            2200,
            {
                "name": "disk_usage_strategy",
                "type": "radiolist",
                "property": "target_disk_usage_strategy",
                "required": True,
                "title": "Disk Usage Strategy",
                "prompt": "Disk usage strategy",
                "dynamic_values": self._get_available_disk_usage_strategies_info,
            },
        )

        self.add_required_config(
            2300,
            {
                "name": "disk_layout_strategy",
                "type": "radiolist",
                "property": "target_disk_layout_strategy",
                "required": True,
                "title": "Disk Layout Strategy",
                "prompt": "Disk layout strategy",
                "dynamic_values": self._get_available_disk_layout_strategies_info,
            },
        )

        self.add_required_config(
            2400,
            {
                "name": "disk_encryption_strategy",
                "type": "radiolist",
                "property": "target_disk_encryption_strategy",
                "required": True,
                "title": "Disk Encryption Strategy",
                "prompt": "Disk encryption strategy",
                "dynamic_values": self._get_available_disk_encryption_strategies_info,
                "dynamic_config_items": self._get_disk_encryption_config_items,
            },
        )

        self.add_required_config(
            2450,
            {
                "name": "rootfs_fstype",
                "type": "radiolist",
                "property": "target_rootfs_fstype",
                "required": True,
                "title": "Root Filesystem Type",
                "prompt": "Root filesystem type",
                "values": [("ext4", "ext4 - Standard filesystem"), ("btrfs", "btrfs - Includes snapshot support")],
            },
        )

        self.add_required_config(
            4100,
            {
                "name": "network_device_naming_strategy",
                "type": "radiolist",
                "property": "target_network_device_naming_strategy",
                "required": True,
                "title": "Network Device Naming Strategy",
                "prompt": "Network device naming strategy",
                "dynamic_values": self._get_available_network_device_naming_strategies_info,
                "dynamic_config_items": self._get_network_device_config_items,
            },
        )

        super().prepare_configure()

    def configure(self) -> None:  # pylint: disable=too-many-branches
        """Configure target."""

        # Run parent configuration
        super().configure()

        # Check if we were booted with a serial console
        if not self.target_serial_terminal:
            with open("/proc/cmdline", "r", encoding="UTF-8") as cmdline_file:
                # Loop with lines
                for line in cmdline_file:
                    # Split options into a list
                    options = line.split()
                    # For now we only support *n8 speed on the first serial port...
                    for option in options:
                        matches = re.match(
                            r"^console=(?P<serial_terminal>ttyS[0-9],(?:115200|38400|9600))",
                            option,
                        )
                        if matches:
                            self.target_serial_terminal = matches.group("serial_terminal")
                            logging.debug(
                                "  - Setting serial console: %s [kernel command line]",
                                self.target_serial_terminal,
                            )
        else:
            logging.debug("  - Setting serial console: %s", self.target_serial_terminal)

        # Get temporarly root mount point
        self.rootfs_path = tempfile.mkdtemp()
        # NK: Make pylance and mypy happy
        rootfs_path = self.rootfs_path
        if not rootfs_path:
            raise IliException("No rootfs_path")
        self.add_temp_path(rootfs_path)

    def post_configure(self) -> None:
        """Post configure."""

        super().post_configure()

        # If the filesystem is btrfs, add btrfs packages
        if self.target_rootfs_fstype == "btrfs":
            self.add_package("btrfs-progs")
            self.add_package("grub-btrfs")
            self.add_package("inotify-tools")
            # Then check if we can disable bootfs
            if self.target_disk_encryption_strategy == "NONE":
                logging.info("  - Disabling 'create_bootfs' for 'btrfs' to support consistent system snapshots")
                self.target_create_bootfs = False
                self.grub_bootfs_prefix = "/@"

    def validate_configuration(self) -> None:  # pylint: disable=too-many-branches
        """Validate the configuration we have."""

        # Propagate upwards
        super().validate_configuration()

        # Make sure we have at least one disk
        if not self.target_disks:
            raise IliException("At least 1 target disk is required to install")

        # Sanity check, make sure disks are actually disks
        logging.debug("  - Target disks")
        for target_disk in self.target_disks:
            # Check if the disk exists
            if not os.path.exists(target_disk):
                raise IliException(f"Target disk '{target_disk}' does not exist")
            # Make sure disk is a block device
            if not pathlib.Path(target_disk).is_block_device():
                raise IliException(f"Target disk '{target_disk}' is not a block device'")
            # Make sure the target disk is available
            if target_disk not in self.available_blockdevices:
                raise IliException(f"Target disk '{target_disk}' is not available for use'")

        # Make sure the disk usage strategy is valid
        logging.debug("  - Disk usage strategy")
        if self.target_disk_usage_strategy not in (
            "ASIS",
            "MDRAID1",
            "MDRAID5",
            "MDRAID6",
            "MDRAID10",
        ):
            raise IliException(f"Disk usage strategy '{self.target_disk_usage_strategy}' is invalid")
        if self.target_disk_usage_strategy not in self._get_available_disk_usage_strategies():
            raise IliException(f"Target disk usage strategy '{self.target_disk_usage_strategy}' is not usable")

        # Make sure the disk layout strategy is valid
        logging.debug("  - Disk layout strategy")
        if self.target_disk_layout_strategy not in ("FLAT", "LVM"):
            raise IliException(f"Disk layout strategy '{self.target_disk_layout_strategy}' is invalid")
        if self.target_disk_layout_strategy not in self._get_available_disk_layout_strategies():
            raise IliException(f"Target disk layout strategy '{self.target_disk_layout_strategy}' is not usable")

        # Make sure the disk layout strategy is valid
        logging.debug("  - Disk encryption strategy")
        if self.target_disk_encryption_strategy not in ("NONE", "LUKS-ON-PART", "LVM-ON-LUKS"):
            raise IliException(f"Disk encryption strategy '{self.target_disk_encryption_strategy}' is invalid")
        if self.target_disk_encryption_strategy not in self._get_available_disk_encryption_strategies():
            raise IliException(f"Target disk encryption strategy '{self.target_disk_encryption_strategy}' is not usable")
        if self.target_disk_encryption_strategy == "LUKS-ON-PART" and self.target_disk_layout_strategy != "FLAT":
            raise IliException(
                f"Target disk encryption strategy '{self.target_disk_encryption_strategy}' invalid for layout "
                "'{self.target_disk_layout_strategy}'"
            )
        if self.target_disk_encryption_strategy == "LVM-ON-LUKS" and self.target_disk_layout_strategy != "LVM":
            raise IliException(
                f"Target disk encryption strategy '{self.target_disk_encryption_strategy}' invalid for layout "
                "'{self.target_disk_layout_strategy}'"
            )
        if self.target_disk_encryption_strategy != "NONE" and not self.target_create_bootfs:
            raise IliException("Disk encryption requires a bootfs to be created.")

        # Automatic rootfs resizing is not supported for MDRAID or LVM
        logging.debug("  - Auto-resize root")
        if self.target_autoresize_root:
            # NK: Make mypy and pylance happy
            if not self.target_disk_usage_strategy:
                raise IliException("target_disk_usage_strategy is not set")
            if self.target_disk_usage_strategy.startswith("MDRAID"):
                raise IliException(
                    f"Automatic resizing of rootfs is not supported with usage strategy '{self.target_disk_usage_strategy}'"
                )
            if self.target_disk_layout_strategy == "LVM":
                raise IliException(
                    f"Automatic resizing of rootfs is not supported with layout strategy '{self.target_disk_layout_strategy}'"
                )
            if self.target_disk_encryption_strategy != "NONE":
                raise IliException("Automatic resizing of rootfs is not supported with encryption")

    def prepare_target(self) -> None:  # noqa: CFQ001 # pylint: disable=too-many-statements,too-many-branches
        """
        Prepare target.

        This must be overridden for each target.
        """

        if self.target_disk_usage_strategy is None:
            raise IliException("target_disk_usage_strategy must be set")
        if self.rootfs_path is None:
            raise IliException("rootfs_path must be set")
        if self.target_disks is None:
            raise IliException("target_disks must be set")

        # Usage strategy ASIS
        if self.target_disk_usage_strategy == "ASIS":
            self._prepare_target_usage_asis()
        # Usage strategy MDRAID
        elif self.target_disk_usage_strategy.startswith("MDRAID"):
            self._prepare_target_usage_mdraid()
            # Add support for MDRAID
            self.add_package("mdadm")
            self.add_grub_efi_module("mdraid1x")
            self.add_mkinitcpio_feature("mdadm")

        if self.main_volume is None:
            raise IliException("main_volume must be set")
        if self.target_rootfs_fstype is None:
            raise IliException("target_rootfs_fstype must be set")

        # Setup the EFI filesystems
        if self.target_create_efi:
            # NK: Make mypy and pylance happy
            if self.target_efifs_fstype is None:
                raise IliException("target_efifs_fstype must be set")

            mount_path_count = 0
            # Loop with the EFI volumes
            for efi_volume in self.efi_volumes:
                # If we have more than one EFI volume we need to add a suffix
                if len(self.efi_volumes) > 1:
                    mount_path_suffix = f"/{mount_path_count}"
                    mount_path_count += 1
                # Else we don't use a suffix
                else:
                    mount_path_suffix = ""
                # Work out the EFI mount path
                efi_mount_path = f"/efi{mount_path_suffix}"

                logging.debug("  - Ading EFI: %s", efi_mount_path)
                # Add EFI filesystem
                self.efifs_list.append(
                    IliFilesystem(
                        source=efi_volume,
                        fstype=self.target_efifs_fstype,
                        mount_root=self.rootfs_path,
                        mount_path=efi_mount_path,
                        options=["-n", "EFI", "-F", "32"],
                    )
                )

        # Add MBR to each device
        for install_disk in self.target_disks:
            logging.debug("  - Adding MBR: %s", install_disk)
            self.add_mbr_device(install_disk)

        # Setup boot filesystem if we are creating one
        if self.target_create_bootfs:
            # NK: Make mypy and pylance happy
            if self.target_bootfs_fstype is None:
                raise IliException("target_bootfs_fstype must be set")
            logging.debug("  - Adding bootfs")
            # If we have encryption enabled, then encrypt the boot volume aswell
            if self.target_disk_encryption_strategy != "NONE":
                # NK: Make mypy and pylance happy
                if self.target_disk_encryption_password is None:
                    raise IliException("target_disk_encryption_password must be set")
                if self.boot_volume is None:
                    raise IliException("boot_volume must be set")
                # Generate UUID for the boot volume
                encrypted_boot_uuid = str(uuid.uuid4())
                luks_name = "boot.encrypted"
                self.create_encrypted_device(
                    luks_name, self.boot_volume, encrypted_boot_uuid, self.target_disk_encryption_password, luks_version=1
                )
                self.encrypted_hints.append(f"{encrypted_boot_uuid}={luks_name}")
                # Make sure we store the bootfs uuid, so we can mount it in grub
                self.encrypted_boot_uuid = encrypted_boot_uuid.replace("-", "")
                # Replace boot volume
                self.boot_volume = f"/dev/mapper/{luks_name}"

            # NK: Make mypy and pylance happy
            if self.boot_volume is None:
                raise IliException("boot_volume must be set")

            self.bootfs = IliFilesystem(
                source=self.boot_volume,
                fstype=self.target_bootfs_fstype,
                mount_root=self.rootfs_path,
                mount_path="/boot",
                options=["-L", "boot"],
            )

        # Create a UUID for our encrypted volume
        encrypted_uuid = ""
        if self.target_disk_encryption_strategy != "NONE":
            # Generate encrypted device UUID
            encrypted_uuid = str(uuid.uuid4())
            # Add encryption to mkinitcpio
            self.add_mkinitcpio_feature("encrypt")

        # If the layout strategy is flat, we just add the root filesystem normally
        root_volume = None
        if self.target_disk_layout_strategy == "FLAT":
            # Check if we need to setup encryption
            if self.target_disk_encryption_strategy == "LUKS-ON-PART":
                # NK: Make mypy and pylance happy
                if self.target_disk_encryption_password is None:
                    raise IliException("target_disk_encryption_password must be set")

                luks_name = "main.encrypted"
                self.create_encrypted_device(luks_name, self.main_volume, encrypted_uuid, self.target_disk_encryption_password)
                self.encrypted_hints.append(f"{encrypted_uuid}={luks_name}")
                # Replace main volume
                self.main_volume = f"/dev/mapper/{luks_name}"

            # Set root volume
            root_volume = self.main_volume

        # If the layout strategy is however LVM, we need to create the PV, VG and LV
        elif self.target_disk_layout_strategy == "LVM":
            # NK: Make sure mypy and pylance are happy
            if self.target_lvm_rootfs_size is None:
                raise IliException("target_lvm_rootfs_size must be set")

            if "MDRAID" in self.target_disk_usage_strategy:
                vg_name = "lvm-raid"
            else:
                vg_name = "main"
            # We always name the root volume "root"
            lv_name = "root"

            # Check if we need to setup encryption...
            if self.target_disk_encryption_strategy == "LVM-ON-LUKS":
                # NK: Make mypy and pylance happy
                if self.target_disk_encryption_password is None:
                    raise IliException("target_disk_encryption_password must be set")

                luks_name = f"{vg_name}.encrypted"
                self.create_encrypted_device(luks_name, self.main_volume, encrypted_uuid, self.target_disk_encryption_password)
                self.encrypted_hints.append(f"{encrypted_uuid}={luks_name}")
                # Replace main volume
                self.main_volume = f"/dev/mapper/{luks_name}"

            logging.debug("  - Setting up LVM")
            # Create LVM physical volume
            self.create_lvm_physical_volume(self.main_volume)
            # Create LVM volume group
            self.create_lvm_volume_group(vg_name, self.main_volume)
            # Create LVM logical volume
            self.create_lvm_logical_volume(lv_name, vg_name, self.target_lvm_rootfs_size)

            # Set root volume
            root_volume = f"/dev/{vg_name}/{lv_name}"
            # Add LVM support
            self.add_package("lvm2")
            self.add_grub_efi_module("lvm")
            self.add_mkinitcpio_feature("lvm")

        # NK: Make sure mypy and pylance are happy
        if root_volume is None:
            raise IliException("root_volume must be set")

        logging.debug("  - Adding rootfs")
        # Add root filesystem onto the root volume we determined above
        self.rootfs = IliFilesystem(
            source=root_volume,
            fstype=self.target_rootfs_fstype,
            mount_root=self.rootfs_path,
            mount_path="/",
            options=["-L", "root"],
        )

        logging.debug("  - Grub bootfs hint: %s", self.grub_bootfs_hint)

    def _prepare_target_usage_asis(self) -> None:
        """
        Prepare the target disk for our ASIS strategy.

        The purpose of this method is to populate the following properties:

        - efi_voulmes: The block devices we're going to be installing our EFI loader on
        - boot_volume: The block device we'll be installing the MBR bootloader on
        - main_volume: The main storage block device
        """

        if self.target_disks is None:
            raise IliException("target_disks must be set")

        # As we're using a single disk as-is, there is only one, grab it
        target_disk = self.target_disks[0]

        logging.debug("  - Preparing disk: ASIS -> %s", target_disk)

        # Get partition prefix for the target disk
        part_prefix = util.disk_partition_prefix(target_disk)

        # Start building partition list
        parts = []
        # Add DOS boot partition
        parts.append(",1M,21686148-6449-6E6F-744E-656564454649")
        # Add EFI partition
        if self.target_create_efi:
            parts.append(f",{self.target_efifs_size},C12A7328-F81F-11D2-BA4B-00A0C93EC93B")
            efi_part_num = len(parts)
            # Add EFI volume
            self.efi_volumes.append(f"{target_disk}{part_prefix}{efi_part_num}")
        # The boot partition is the next partition, either the bootfs or the rootfs
        boot_part_num = len(parts) + 1
        # Add bootfs
        if self.target_create_bootfs:
            parts.append(f",{self.target_bootfs_size},BC13C2FF-59E6-4262-A352-B275FD6F7172")
            self.boot_volume = f"{target_disk}{part_prefix}{boot_part_num}"
        # Add main partition
        parts.append(",,4F68BCE3-E8CD-4DB1-96E7-FBCAF984B709")
        main_part_num = len(parts)

        # Write out partitions, we only have one disk so we use the first one
        logging.debug("  - Writing partition table: %s", target_disk)
        self.write_partitions(target_disk, "gpt", parts)

        # Set the grub hint and boot volume
        self.grub_bootfs_hint = f"hd0,gpt{boot_part_num}"
        # Set main volume
        self.main_volume = f"{target_disk}{part_prefix}{main_part_num}"

    def _prepare_target_usage_mdraid(self) -> None:
        """
        Prepare the target disk for our ASIS strategy.

        The purpose of this method is to populate the following properties:

        - efi_voulmes: The block devices we're going to be installing our EFI loader on
        - boot_volume: The block device we'll be installing the MBR bootloader on
        - main_volume: The main storage block device
        """

        if self.target_disks is None:
            raise IliException("target_disks must be set")
        if self.target_disk_usage_strategy is None:
            raise IliException("target_disk_usage_strategy must be set")

        # Components of the various MDRAID devices we'll be creating
        boot_volume_components = []
        main_volume_components = []

        logging.debug(
            "  - Preparing disk: %s -> %s",
            self.target_disk_usage_strategy,
            " ,".join(self.target_disks),
        )

        # Loop with each RAID disk
        boot_volume_uuid = None
        for raid_disk in self.target_disks:
            # Get partition prefix for the target disk
            part_prefix = util.disk_partition_prefix(raid_disk)

            parts = []
            # Add DOS boot partition
            parts.append(",1M,21686148-6449-6E6F-744E-656564454649")
            # Add EFI partition
            efi_part_num: Optional[int] = None
            if self.target_create_efi:
                parts.append(f",{self.target_efifs_size},C12A7328-F81F-11D2-BA4B-00A0C93EC93B")
                efi_part_num = len(parts)
                # Add EFI volume
                self.efi_volumes.append(f"{raid_disk}{part_prefix}{efi_part_num}")
            # Check if we're adding a boot partition
            boot_part_num: Optional[int] = None
            if self.target_create_bootfs:
                parts.append(f",{self.target_bootfs_size},A19D880F-05FC-4D3B-A006-743F0F84911E")
                boot_part_num = len(parts)
                # Add bootfs partition to list
                boot_volume_components.append(f"{raid_disk}{part_prefix}{boot_part_num}")
            # Add rest of device partition
            parts.append(",,A19D880F-05FC-4D3B-A006-743F0F84911E")
            main_part_num = len(parts)

            # Add main partition to list
            main_volume_components.append(f"{raid_disk}{part_prefix}{main_part_num}")

            # Write out partitions
            logging.debug(
                "  - %s component partitions: %s",
                self.target_disk_usage_strategy,
                raid_disk,
            )
            self.write_partitions(raid_disk, "gpt", parts)

        # Check if we're creating a RAID1 boot device
        if self.target_create_bootfs:
            boot_volume_uuid = util.md_uuid()
            self.boot_volume = "/dev/md/boot"
            logging.debug("  - MDRAID1 boot: %s", ", ".join(boot_volume_components))
            # Once we've partitioned the disks, add a RAID1 across all boot partitions
            # NK: Make pylance and mypy happy
            boot_volume = self.boot_volume
            if not boot_volume:
                raise IliException("boot_volume not set")
            self.create_mdraid(boot_volume, boot_volume_uuid, 1, boot_volume_components)

        # Create the main volume
        main_volume_uuid = util.md_uuid()
        raid_level = int(self.target_disk_usage_strategy[6:])
        self.main_volume = "/dev/md/main"
        logging.debug(
            "  - %s main: %s",
            self.target_disk_usage_strategy,
            ", ".join(main_volume_components),
        )
        main_volume = self.main_volume
        if not main_volume:
            raise IliException("main_volume not set")
        self.create_mdraid(main_volume, main_volume_uuid, raid_level, main_volume_components)

        # Set the grub bootfs hint
        if self.target_create_bootfs:
            # NK: Make mypy and pylance happy
            if boot_volume_uuid is None:
                raise IliException("boot_volume_uuid is not set")
            self.grub_bootfs_hint = "mduuid/" + boot_volume_uuid.replace(":", "")
        else:
            self.grub_bootfs_hint = "mduuid/" + main_volume_uuid.replace(":", "")

    def setup_target_mdadm_conf(self) -> None:
        """Set up /etc/mdadm.conf if we're using software RAID."""

        if self.rootfs is None:
            raise IliException("rootfs must be set")
        if self.rootfs.mount_point is None:
            raise IliException("rootfs mount_point must be set")
        if self.target_disk_usage_strategy is None:
            raise IliException("target_disk_usage_strategy must be set")

        # If we're not using software raid, we can just return here
        if not self.target_disk_usage_strategy.startswith("MDRAID"):
            logging.info("No need to create mdmadm.conf as MDRAID is not used")
            return

        logging.debug("  - Setting up mdadm.conf")

        # Grab the disk array details
        mdadm_output = self.run_target(["/usr/bin/mdadm", "--examine", "--scan"])

        # Write out our mdadm configuration
        mdadm_conf_path = f"{self.rootfs.mount_point}/etc/mdadm.conf"
        try:
            with open(mdadm_conf_path, "a", encoding="UTF-8") as mdadm_conf_file:
                mdadm_conf_file.write("\n")
                mdadm_conf_file.write("\n".join(mdadm_output))
                mdadm_conf_file.write("\n")
                mdadm_conf_file.write("MAILADDR root@localhost\n")
        except OSError as err:
            raise IliException(f"Unable to open file '{mdadm_conf_path}' for writing: {err}") from None

    def finalize_target(self) -> None:
        """Finalize target."""

        if self.target_disks is None:
            raise IliException("target_disks must be set")

        logging.info("Installation complete: %s", ", ".join(self.target_disks))

    def _get_available_blockdevices_info(self) -> List[Tuple[str, str]]:  # pylint: disable=useless-option-value
        """Return available block devices."""

        # Open subprocess to lsblk
        lsblk_output = util.run_host(
            [  # nosec
                "/usr/bin/lsblk",
                "--list",
                "--noheadings",
                "--output",
                "PATH,NAME,TYPE,SIZE",
                "--sort",
                "NAME",
                "--nodeps",
                "--include",
                "8,179,254,259",
                "--json",
            ]
        )

        # Make sure we got something back
        if not lsblk_output:
            return []

        # Grab devices from the JSON output
        json_decoded = json.loads("".join(lsblk_output))

        # Return available block devices
        return [(x["path"], f"{x['name']} [{x['size']}]") for x in json_decoded["blockdevices"]]

    def _get_available_network_devices(self) -> Dict[str, Dict[str, str]]:  # pylint: disable=useless-option-value
        """Return available network devices on the system."""

        class_dir = "/sys/class/net"

        network_devices: Dict[str, Dict[str, str]] = {}
        # List the directory contents so we can get the device names
        try:
            network_device_list = os.listdir(class_dir)
        except OSError as err:
            raise IliException(f"Failed listdir() on '{class_dir}': {err}") from None

        # Loop with each device name and build its associated info
        for device_name in network_device_list:
            # If it does not start with an "e", then its not ethernet, right?
            if not device_name.startswith("e"):
                continue
            # Path of the device directory
            device_dir_path = f"{class_dir}/{device_name}"
            # Initialize device info
            network_devices[device_name] = {}
            # Grab symlink path
            network_devices[device_name]["symlink"] = os.readlink(device_dir_path)
            # Open the hardware address file and read in the MAC address
            device_address_path = f"{device_dir_path}/address"
            try:
                with open(device_address_path, "r", encoding="UTF-8") as address_file:
                    # Read in MAC and strip the newline
                    network_devices[device_name]["mac"] = address_file.read().rstrip()
            except OSError as err:
                raise IliException(f"Failed to open '{device_address_path}' for reading: {err}") from None

        return network_devices

    def _get_available_disk_usage_strategies_info(self) -> List[Tuple[str, str]]:
        """Return available disk usage strategies info."""

        # Make pylint happy
        if not self.target_disks:
            return []

        disk_count = len(self.target_disks)

        strategies = []

        # If we only have one disk, then the only option for us is AS-IS
        if disk_count == 1:
            strategies.append(("ASIS", "Use disk AS-IS"))

        # If we have more than 1 disk, then RAID1 will work
        if disk_count > 1:
            strategies.append(("MDRAID1", "Setup MDRAID1"))

        # If we have 3 or more disks then RAID5 will work
        if disk_count >= 3:
            strategies.append(("MDRAID5", "Setup MDRAID5"))

        # If we have 4 or more disks then RAID6 will work
        if disk_count >= 4:
            strategies.append(("MDRAID6", "Setup MDRAID6"))

        # If we have 4 or more disks, and they are in multiples of 2, then RAID10 will work
        if (disk_count >= 4) & (disk_count % 2 == 0):
            strategies.append(("MDRAID10", "Setup MDRAID10"))

        return strategies

    def _get_available_disk_usage_strategies(self) -> List[str]:
        """Return available disk usage strategies."""
        return [x[0] for x in self._get_available_disk_usage_strategies_info()]

    def _get_available_disk_layout_strategies_info(self) -> List[Tuple[str, str]]:  # pylint: disable=useless-option-value
        """Return available disk layout strategies info."""

        strategies = []
        strategies.append(("FLAT", "Flat partitioning"))
        strategies.append(("LVM", "Use LVM volumes"))

        return strategies

    def _get_available_disk_layout_strategies(self) -> List[str]:
        """Return available disk layout strategies."""
        return [x[0] for x in self._get_available_disk_layout_strategies_info()]

    def _get_available_disk_encryption_strategies_info(self) -> List[Tuple[str, str]]:
        """Return available disk encryption strategies info."""

        strategies = []
        strategies.append(("NONE", "No encryption"))
        if self.target_disk_layout_strategy == "FLAT":
            strategies.append(("LUKS-ON-PART", "Use LUKS on PART"))
        elif self.target_disk_layout_strategy == "LVM":
            strategies.append(("LVM-ON-LUKS", "Use LVM on LUKS"))

        return strategies

    def _get_available_disk_encryption_strategies(self) -> List[str]:
        """Return available disk encryption strategies."""
        return [x[0] for x in self._get_available_disk_encryption_strategies_info()]

    def _get_available_network_device_naming_strategies_info(
        self,
    ) -> List[Tuple[str, str]]:
        """Return available network device naming strategies info."""

        strategies = []

        # Only return options if networking is enabled
        if self.target_networking:
            strategies.append(("ASIS", "Use device names as-is"))
            strategies.append(("AUTO-SEQ", "Rename devices to eXpY"))

        return strategies

    def _get_available_network_device_naming_strategies(self) -> List[str]:
        """Return available network device naming strategies."""
        return [x[0] for x in self._get_available_network_device_naming_strategies_info()]

    def __populate_network_devices_asis(self) -> None:
        """Populate our network device list AS-IS."""

        # Add network devices as is to the list of devices we'll be asking configuration for
        for name, device in self.available_network_devices.items():
            self.target_network_devices.append(name)
            self.target_network_config[f"network_device_{name}_mac"] = device["mac"]

    def __populate_network_devices_autoseq(self) -> None:
        """Populate our network device list using AUTO-SEQ."""

        # Sort devices on name length, lower lengths are more than likely onboard interfaces
        devices_sorted_len: Dict[int, List[str]] = {}
        for device_name in sorted(self.available_network_devices, key=len):
            name_length = len(device_name)
            # If we don't have this length yet, initialize the device list
            if name_length not in devices_sorted_len:
                devices_sorted_len[name_length] = []
            # Add device
            devices_sorted_len[name_length].append(device_name)

        # Start with device 0 port 0
        device_num = 0
        port_num = 0
        device_position_prev = ""
        # Loop with device name lengths
        for _, device_list in sorted(devices_sorted_len.items()):
            # Loop with each device that has this name length
            for device_name in device_list:
                # Grab bus location
                # eg. ../../devices/pci0000:00/0000:00:1c.0/0000:08:00.0/net/enp8s0f0
                device_position = self.available_network_devices[device_name]["symlink"].split("/")[4]
                # If we've not seen any devices, set the previous device to this one
                if not device_position_prev:
                    device_position_prev = device_position
                # Check if this network device is in a different position
                if device_position != device_position_prev:
                    device_num += 1
                    port_num = 0
                    device_position_prev = device_position
                # Save the new name for the device and its MAC address
                new_device_name = f"e{device_num}p{port_num}"
                self.add_network_device(new_device_name)
                self.target_network_config[f"network_device_{new_device_name}_mac"] = self.available_network_devices[device_name][
                    "mac"
                ]
                # Bump the port number
                port_num += 1

    def _get_network_device_config_items(self, value: str) -> List[Dict[str, Any]]:
        """Return network type configuration for each of the network devices."""

        # Check if we're re-naming our network devices
        if self.target_network_device_naming_strategy == "ASIS":
            self.__populate_network_devices_asis()
        elif self.target_network_device_naming_strategy == "AUTO-SEQ":
            self.__populate_network_devices_autoseq()

        return super()._get_network_device_config_items(value)

    def _get_disk_encryption_config_items(self, value: str) -> List[Dict[str, Any]]:
        """Return disk encryption configuration items."""

        # Make pylint happy
        if value is None:
            return []

        config_items = []

        # If we have a disk encryption strategy, then we need to ask for the password
        if self.target_disk_encryption_strategy != "NONE":
            config_items.append(
                {
                    "name": "disk_encryption_password",
                    "type": "password",
                    "property": "target_disk_encryption_password",
                    "title": "Disk encryption password",
                    "prompt": "Disk encryption password",
                },
            )

        return config_items

    @property
    def available_blockdevices_info(self) -> List[Tuple[str, str]]:
        """Return available block devices info."""
        return self._available_blockdevices_info

    @property
    def available_blockdevices(self) -> List[str]:
        """Return available block devices list."""
        return [x[0] for x in self.available_blockdevices_info]

    @property
    def available_network_devices(self) -> Dict[str, Dict[str, str]]:
        """Return available network devices."""
        return self._available_network_devices

    @property
    def target_disks(self) -> Optional[List[str]]:
        """List of target disks."""
        return self._target_disks

    @target_disks.setter
    def target_disks(self, disk_list: List[str]) -> None:
        """Set the target disks to use."""
        # Split the list provided on comma's and add them
        self._target_disks = []
        for item in disk_list:
            for target_disk in item.split(","):
                self._target_disks.append(target_disk)

    @property
    def target_disk_usage_strategy(self) -> Optional[str]:
        """Target disk usage strategy."""
        return self._target_disk_usage_strategy

    @target_disk_usage_strategy.setter
    def target_disk_usage_strategy(self, disk_usage_strategy: str) -> None:
        """Set target disk usage strategy."""
        self._target_disk_usage_strategy = disk_usage_strategy

    @property
    def target_disk_layout_strategy(self) -> Optional[str]:
        """Target disk layout strategy."""
        return self._target_disk_layout_strategy

    @target_disk_layout_strategy.setter
    def target_disk_layout_strategy(self, disk_layout_strategy: str) -> None:
        """Set target disk layout strategy."""
        self._target_disk_layout_strategy = disk_layout_strategy

    @property
    def target_disk_encryption_strategy(self) -> Optional[str]:
        """Target disk encryption strategy."""
        return self._target_disk_encryption_strategy

    @target_disk_encryption_strategy.setter
    def target_disk_encryption_strategy(self, disk_encryption_strategy: str) -> None:
        """Set target disk encryption strategy."""
        self._target_disk_encryption_strategy = disk_encryption_strategy

    @property
    def target_network_device_naming_strategy(self) -> Optional[str]:
        """Target network device naming strategy."""
        return self._target_network_device_naming_strategy

    @target_network_device_naming_strategy.setter
    def target_network_device_naming_strategy(self, network_device_naming_strategy: str) -> None:
        """Set target network device naming strategy."""
        self._target_network_device_naming_strategy = network_device_naming_strategy

    @property
    def target_disk_encryption_password(self) -> Optional[str]:
        """Disk encryption password."""
        return self._target_disk_encryption_password

    @target_disk_encryption_password.setter
    def target_disk_encryption_password(self, disk_encryption_password: str) -> None:
        """Disk encryption password."""
        self._target_disk_encryption_password = disk_encryption_password

    @property
    def target_lvm_rootfs_size(self) -> Optional[str]:
        """In terms of LVM, this is the root filesystem size to create."""
        return self._target_lvm_rootfs_size

    @target_lvm_rootfs_size.setter
    def target_lvm_rootfs_size(self, rootfs_size: str) -> None:
        """In terms of LVM, set the root filesystem size to create."""
        self._target_lvm_rootfs_size = rootfs_size

    @property
    def rootfs_path(self) -> Optional[str]:
        """Temporary rootfs path."""
        return self._rootfs_path

    @rootfs_path.setter
    def rootfs_path(self, rootfs_path: str) -> None:
        """Set temporary rootfs path."""
        self._rootfs_path = rootfs_path

    @property
    def efi_volumes(self) -> List[str]:
        """Return the list of EFI volumes."""
        return self._efi_volumes

    @property
    def boot_volume(self) -> Optional[str]:
        """Return the boot volume."""
        return self._boot_volume

    @boot_volume.setter
    def boot_volume(self, boot_volume: str) -> None:
        """Set the boot volume."""
        self._boot_volume = boot_volume

    @property
    def main_volume(self) -> Optional[str]:
        """Return the main volume."""
        return self._main_volume

    @main_volume.setter
    def main_volume(self, main_volume: str) -> None:
        """Set the main volume."""
        self._main_volume = main_volume
